#!/usr/bin/env python
#
# Agalma - Tools for processing gene sequence data and automating workflows
# Copyright (c) 2012 Brown University. All rights reserved.
# 
# This file is part of Agalma.
# 
# Agalma is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Agalma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Agalma.  If not, see <http://www.gnu.org/licenses/>.

"""
Assembles reads into transcripts, processes the assembly, and generates 
assembly diagnostics. Read pairs are first filtered at a more stringent 
mean quality threshold. Assemblies are then performed over a range of 
data subset sizes, which provides an indication of how sequencing effort 
impacts assembly results.
"""

import os

from biolite import diagnostics
from biolite import report
from biolite import utils
from biolite import workflows
from biolite import wrappers

from biolite.pipeline import IlluminaPipeline

pipe = IlluminaPipeline('assemble', __doc__)

# Command-line arguments.

pipe.add_argument('quality', short='q', type=int, metavar='MIN',
	default=33, help="""
	Filter out reads that have a mean quality < MIN.""")

pipe.add_argument('kmers', short='k', type=int, nargs='+', metavar='K',
	default=(31, 35, 41, 47, 53), help="""
	A list of kmer sizes for Oases-M assembly.""")

pipe.add_argument('subsets', short='s', type=int, nargs='+', metavar='N',
	default=None, help="""
	A list of cumulative subsets of N reads to assemble.""")

pipe.add_argument('min_length', short='l', type=int, metavar='L',
	default=300, help="""
	Only keep transcripts longer than L bases.""")

# Pipeline stages.

@pipe.stage
def quality_filter(id, data, quality, subsets):
	"""Filter out low-quality reads"""

	hiqual = []

	for i,_ in enumerate(data[-1]):
		hiqual.append(id + '.hiqual.%d.fastq' % (i+1))

	if len(data[-1]) == 1:
		# Single end data probably needs adapter and base content filtering.
		wrappers.FilterIllumina(data[-1], hiqual, quality=quality)
	else:
		# Paired end probably ran through earlier agalma pipelines, so don't
		# redo the adapater and base content filtering, but do reformat the
		# separator to '/' from CASAVA 1.6
		wrappers.FilterIllumina(
					data[-1], hiqual, quality=quality,
					adapters=False, bases=False, sep='/')

	# Parse the line count from the output.
	entity = "%s.filter_illumina" % diagnostics.get_entity()
	nreads = int(diagnostics.local_lookup(entity)['pairs_kept'])

	# Find the max subset that is less than the full set of reads.
	if subsets:
		subsets = map(int, subsets)
		subsets = filter(lambda x: x < nreads, subsets)
		subsets.append(nreads)
	else:
		subsets = [nreads]

	diagnostics.log('subsets', subsets)

	data.append(hiqual)

	assemblies = []
	assemblies_full = []

	ingest('subsets', 'assemblies', 'assemblies_full')

@pipe.stage
def oases_merge(assemblies, assemblies_full, data, subsets, kmers, min_length, outdir):
	"""Oases-M assemblies of cumulative subsets of reads"""
	for n in subsets:
		assembly = 'assembly_%d_oases-m' % n
		utils.safe_mkdir(assembly)
		diagnostics.prefix.append(str(n))
		# Accumulate more reads for each cumulative subset.
		if (n != subsets[-1]):
			subset = []
			for i,d in enumerate(data[-1]):
				subset.append('subset.%d.fastq' % (i+1))
				utils.head_to_file(d, subset[-1], 4*n, 'w')
		else:
			subset = data[-1]
		# Assemble, sweeping over kmer sizes.
		contigs = os.path.join(outdir, assembly + '.fa')
		workflows.oases_merge_assembly(
			subset, contigs, 31, kmers, min_length, workdir=assembly)
		assemblies.append(contigs)
		diagnostics.log_path(contigs)
		diagnostics.prefix.pop()

	assemblies_full.append(contigs)

@pipe.stage
def trinity(assemblies, assemblies_full, data, subsets, min_length, outdir):
	"""Trinity assemblies of cumulative subsets of reads"""
	subset1 = 'subset.1.fastq'
	subset2 = 'subset.2.fastq'
	for n in subsets:
		assembly = 'assembly_%d_trinity' % n
		utils.safe_mkdir(assembly)
		diagnostics.prefix.append(str(n))
		# Accumulate more reads for each cumulative subset.
		if (n != subsets[-1]):
			subset = []
			for i,d in enumerate(data[-1]):
				subset.append('subset.%d.fastq' % (i+1))
				utils.head_to_file(d, subset[-1], 4*n, 'w')
		else:
			subset = data[-1]
		contigs = os.path.join(outdir, assembly + '.fa')
		workflows.trinity_assembly(contigs, subset, assembly, min_length)
		assemblies.append(contigs)
		diagnostics.log_path(contigs)
		diagnostics.prefix.pop()

	assemblies_full.append(contigs)

@pipe.stage
def sync(data, assemblies, assemblies_full):
	"""Sync up after all assemblies are finished, and output the list of assemblies"""
	finish('data', 'assemblies', 'assemblies_full')

if __name__ == "__main__":
	# Run the pipeline.
	pipe.run()
	# Push the local diagnostics to the global database.
	diagnostics.merge()

# Report. #
Field = report.Field
filter_schema = (
	Field('pairs', "Read pairs examined", int, '{:,d}'),
	Field('pairs_kept', "Read pairs kept", int, '{:,d}'),
	Field('percent_kept', "Percent kept", float, '{:.1f}%'),
	Field('quality', "Illumina quality threshold", int, '{:,d}'),
	Field('reject_adapters', "Adapter fails", int, '{:,d}'),
	Field('reject_quality', "Quality fails", int, '{:,d}'),
	Field('reject_content', "Base composition fails", int, '{:,d}'))

class Report(report.BaseReport):
	def init(self):
		self.name = pipe.name
		# Lookups
		self.lookup('filter', 'assemble.quality_filter.filter_illumina')
		self.percent('filter', 'percent_kept', 'pairs_kept', 'pairs')
		# Generators
		self.generator(self.filter_table)

	def filter_table(self):
		if 'filter' in self.data:
			# Pull the quality treshold out of the command arguments.
			self.data.filter['quality'] = self.extract_arg(
								'assemble.quality_filter.filter_illumina', '-q')
			html = [self.header("Illumina Filtering")]
			html += self.summarize(filter_schema, 'filter')
			return html

