# Agalma - Tools for processing gene sequence data and automating workflows
# Copyright (c) 2012 Brown University. All rights reserved.
# 
# This file is part of Agalma.
# 
# Agalma is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Agalma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Agalma.  If not, see <http://www.gnu.org/licenses/>.

"""
Provides an interface to the underlying SQLite database that stores the
Agalma **sequences**, **rpsblast**, and **homology** tables.
"""

import os
import sqlite3

from biolite import utils
from biolite.config import get_resource

_db = None

def _create_sql(name, schema, index):
	sql = [" CREATE TABLE %s (" % name]
	sql += ["   %s %s," % s for s in schema[:-1]]
	sql += ["   %s %s);" % schema[-1]]
	sql += [" CREATE INDEX {0}_{1} ON {0}({1});".format(name, i) for i in index]
	return '\n'.join(sql)

sequences_schema = (
	('sequence_id', 'INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL'),  
	('run_id', 'INTEGER'),			# Foreign key to runs table, refers to the run_id of the process that populates the database
	('catalog_id', 'VARCHAR(256)'),			# Foreign key to catalog.id, could be obtained through join with run table but including it here makes things more robust
	('locus', 'INTEGER'),			# Parsed from the header, eg 3 from Locus_3_Transcript_20
	('transcript', 'INTEGER'),			# Parsed from the header, eg 20 from Locus_3_Transcript_20
	('confidence', 'FLOAT'),			# Parsed from the header
	('rpkm', 'FLOAT'),			# Parsed from the header, null for non-exemplars
	('exemplar', 'INTEGER'),			# 0 if not an exemplar, 1 if an exemplar (all rows will be 1 if loading from the exemplar file)
	('nucleotide_seq', 'TEXT'),			# Nucleotide sequence from either exemplars or transcripts file (which is loaded will be up to the user)
	('nucleotide_quality', 'TEXT'),			# Quality string corresponding to nucleotide_seq, NULL for most assemblers (including oases)
	('protein_seq', 'TEXT'),			# Parsed from the prot4EST output, for now will be null for non-exemplars since they aren't translated. 
	('translation_method', 'VARCHAR(256)'),			# A string with the program and method, eg "prot4EST SIMILARITY", only applies when protein_seq is not NULL
	('organelle', 'VARCHAR(256)'),			# Either nuclear or mitochondrial, parsed from prot4EST output. NULL for non exemplars.
	('blast_hit', 'TEXT'),			# The name, gi, and evalue of the top blastx hit, could be parsed from the nucleotide fasta header or the blastx report. Provides a human readable label. NULL for non-exemplars
	('nucleotide_header', 'TEXT'),	# The header in the nucleotide fasta from which the sequence was loaded
	('protein_header', 'TEXT'),	# The header in the protein fasta from which the sequence was loaded
	('note', 'TEXT')			# Useful for user specified information')
)

sequences_index = ('run_id', 'catalog_id', 'locus', 'transcript')
sequences_sql = _create_sql('sequences', sequences_schema, sequences_index)

rpsblast_schema = (
	('run_id', 'INTEGER'),			# Foreign key to runs table, refers to the run_id of the process that populates the database
	('catalog_id', 'VARCHAR(256)'),			# query - Foreign key to catalog.id, could be obtained through join with run table but including it here makes things more robust
	('locus', 'INTEGER'),			# query - Parsed from the header, eg 3 from Locus_3_Transcript_20
	('transcript', 'INTEGER'),			# query - Parsed from the header, eg 20 from Locus_3_Transcript_20
	('hit_id', 'VARCHAR(256)'), 		# subject - accession number, parsed from <Hit_id> field of blast report
	('mask', 'TEXT'),		# mask showing where hits are on query
	('evalue', 'FLOAT'),
	('score', 'FLOAT')
)

rpsblast_index = ('run_id', 'catalog_id', 'locus', 'transcript')
rpsblast_sql = _create_sql('rpsblast', rpsblast_schema, rpsblast_index)

homology_schema = (
	('run_id', 'INTEGER'),			# Foreign key to runs table, refers to the run_id of the process that assigned homology and populated this table
	('component_id', 'INTEGER'),	# The id assigned to the component during graph processing
	('sequence_id', 'INTEGER'),		# Foreign key to sequences table, identifies the sequence assigned to a particular component
)

homology_index = ('run_id', 'component_id')
homology_sql = _create_sql('homology', homology_schema, homology_index)

def _create_db():
	global _db
	_db.executescript(sequences_sql + rpsblast_sql + homology_sql)

def connect(*args, **kwargs):
	"""
	Establish a gobal database connection and set the `execute` function.
	"""
	global _db, execute
	if _db is None:
		path = get_resource('agalma_database')
		exists = os.path.isfile(path)
		if not exists:
			utils.safe_mkdir(os.path.dirname(path))
		_db = sqlite3.connect(path, timeout=60.0, isolation_level=None)
		if not exists:
			_create_db()
		execute = _db.execute
	if len(args):
		return execute(*args, **kwargs)

def disconnect():
	"""
	Close the global database connection, set it to None, and set the
	`execute` function to auto-reconnect.
	"""
	global _db, execute
	if _db is not None:
		_db.close()
	_db = None
	execute = connect

execute = connect

