#!/usr/bin/env python
#
# Agalma - Tools for processing gene sequence data and automating workflows
# Copyright (c) 2012 Brown University. All rights reserved.
# 
# This file is part of Agalma.
# 
# Agalma is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Agalma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Agalma.  If not, see <http://www.gnu.org/licenses/>.

"""
Load

This pipeline parses assembled sequences and loads them into agalma's database.
It requires:

* A nucleotide fasta file.
* The nucleotide sequence name cannot have any spaces or | in it. Any
  characters after the first | are considered as annotations, and the
  annotations may have spaces.
* The nucleotide file has been generated by earlier Agalma pipelines. The
  nucleotide fasta headers are in Oases format.  The nucleotide header base can
  be followed with additional | delimited annotations. This mode supports
  having multiple transcripts per locus, but assumes that there is only one
  translation per locus (this translation is for the single exemplar sequence).
* If the --generic flag is used, you can import sequences without the above
  restrictions. The locus name will be sequence name up to the first space.
"""

import os
import re

from Bio import SeqIO
from collections import OrderedDict, namedtuple

import database

from biolite import catalog
from biolite import config
from biolite import diagnostics
from biolite import report
from biolite import utils
from biolite import workflows

from biolite.pipeline import Pipeline

pipe = Pipeline('load', __doc__)

# Command-line arguments.

pipe.add_argument('nuc', short='n', default=None, help="""
	Path to input nucleotide fasta file""")
pipe.add_argument('generic', action='store_true', default=False, help="""
	Nucleotide fasta file has a generic (non-agalma) header format: the locus
	will be parsed as the header up to the first space character found""")

# Pipeline stages.

@pipe.stage
def setup_path(id, previous, nuc):
	"""Determine the path to the nucleotide input"""
	if nuc:
		nuc = os.path.abspath(nuc)
	elif previous:
		# Lookup output path from a previous assembly run.
		values = diagnostics.lookup_prev_run(id, previous)
		try:
			assembly_paths = diagnostics.str2list(values['assembly_paths'])
		except KeyError:
			utils.die("could not find previous 'assemble' run for '%s'" % id)
		nuc = assembly_paths[-1][-1]
		utils.info("using last assembly '%s'" % nuc)
	else:
		record = catalog.select(id)
		if record and len(record.paths):
			nuc = catalog.split_paths(record.paths)[0]
			utils.info("using first path in the catalog: '%s'" % nuc)
		else:
			utils.die("no nucleotide file or previous run specified")

	diagnostics.log_path(nuc, 'nuc')
	ingest('nuc')


@pipe.stage
def load_sequences(id, _run_id, nuc, generic):
	"""Read the nucleotide sequences from the FASTA file"""

	# Store sequences as lists of transcripts keyed by locus
	sequences = OrderedDict()

	# RE for validating sequence string
	alphabet = re.compile(r"^[atcgnATCGN]+$")

	Sequence = namedtuple(
				'Sequence',
				"seq header confidence rpkm exemplar blast_hit")

	for i, record in enumerate(SeqIO.parse(open(nuc), 'fasta')):

		seq = str(record.seq)

		# Validate sequence.
		if not alphabet.match(seq):
			utils.info("bad sequence for FASTA record %d" % (i+1))
			continue

		header = record.description

		# Unpack the nucleotide header.

		# Example Agalma header:
		# >Locus_6_Transcript_4/10_Confidence_0.652_Length_2824|RPKM|273.53|BlastHit|gi|313118273|sp|Q9SXD5.2|GSXL3_ARATH RecName: Full=Flavin-containing monooxygenase FMO GS-OX-like 3; AltName: Full=Flavin-monooxygenase glucosinolate S-oxygenase-like 3|7.25911e-38

		# Example generic header (with BlastHit annotation):

		fields = header.split('|')
		annot_fields = fields[1:]

		# Parse out BlastHit annotation

		if 'BlastHit' in annot_fields:
			i = annot_fields.index('BlastHit') + 1
			blast_hit = '|'.join(annot_fields[i:])
			blast_hit = blast_hit.replace("'", "") # Sanitize the string
		else:
			blast_hit = None

		# Generic headers don't have locus/transcript info or RPKM ...
		if generic:

			locus = fields[0].partition(' ')[0]

			if not locus:
				utils.die("couldn't parse header '%s'" % header)
			elif locus in sequences:
				utils.die("duplicate header found for '%s'" % locus)

			sequences[locus] = (Sequence(seq, header, 0.0, -1 , 1, blast_hit),)

		# ... Agalma headers do
		else:	

			# Unpacking is a two stage process, first unpacking the
			# oases-style data that come first (through length), then the
			# data that were added during annotation

			oases_fields = workflows.unpack_oases_header(fields[0])

			# Trinity loci have the form N.M for ints N and M
			# Oases loci are just an int N
			# We leave it as a string for now
			locus = oases_fields.locus

			# By default, use the first (or only) transcript as the
			# exemplar
			if locus in sequences:
				exemplar = 0
			else:
				exemplar = 1
				sequences[locus] = list()
			
			confidence = float(oases_fields.confidence)

			if 'RPKM' in annot_fields:
				rpkm = float(annot_fields[annot_fields.index('RPKM') + 1])
			else:
				rpkm = -1

			sequences[locus].append(Sequence(
				seq, header, confidence, rpkm, exemplar, blast_hit))


	# Bulk insert into the sequences table
	database.execute('BEGIN')

	# Clear any previous/failed loads
	database.execute("DELETE FROM sequences WHERE run_id=?;", (_run_id,))

	sql = """
		INSERT INTO sequences (
			run_id, catalog_id, locus, transcript, confidence, rpkm,
			exemplar, nucleotide_seq, protein_seq, translation_method,
			organelle, blast_hit, nucleotide_header, protein_header
		) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?);"""

	# Agalma doesn't currently provide protein translations, so these
	# fields are blank.
	protein_seq = ''
	protein_header = ''
	translation_method = ''
	protein_orientation = ''
	organelle = ''

	# Loop over loci...
	locus = 0
	for key in sequences:
		locus += 1
		# And transcripts...
		transcript = 0
		for seq in sequences[key]:
			transcript += 1
			database.execute(sql, (
				_run_id, id, locus, transcript, seq.confidence,
				seq.rpkm, seq.exemplar, seq.seq, protein_seq,
				translation_method, organelle, seq.blast_hit,
				seq.header, protein_header))
	
	# Finish up
	database.execute("COMMIT")


if __name__ == "__main__":
	# Run the pipeline.
	pipe.run()
	# Push the local diagnostics to the global database.
	diagnostics.merge()

# Report. #
class Report(report.BaseReport):
	def init(self):
		self.name = pipe.name
		# Lookups
		self.lookup('outdir', diagnostics.INIT, 'outdir')
		# Generators
