#!/usr/bin/env python
#
# Agalma - Tools for processing gene sequence data and automating workflows
# Copyright (c) 2012 Brown University. All rights reserved.
# 
# This file is part of Agalma.
# 
# Agalma is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Agalma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Agalma.  If not, see <http://www.gnu.org/licenses/>.

import os
import argparse
import codecs

from string import Template
from collections import OrderedDict

from biolite import catalog
from biolite import diagnostics
from biolite import report
from biolite import utils

import sanitize
import insert_size
import remove_rrna
import assemble
import postassemble
import homologize
import multalign

def profile_pipeline(run_id):
	values = diagnostics.lookup(run_id, "%s.profile" % diagnostics.EXIT)
	profile = list()
	for field in report.profile_schema[2:]:
		op = report.profile_aggregators.get(field.key, sum).__name__
		try:
			profile.append("%-19s: %12s [%s]" % (
				field.title,
				field.format.format(field.type(values[field.key])),
				op))
		except KeyError:
			profile.append("%-19s: %12s [%s]" % (field.title, '-', op))
	return '\n'.join(profile)

def catalog_header(record):
	row = "<tr><td><b>%s</b></td><td>%s</td></tr>"
	html = [
		"<table>",
		row % (catalog.fields[0], record[0])]
	html += [row % x for x in zip(catalog.fields[3:-1], record[3:-1])]
	html.append("</table>")
	return html

def report_runs(id, outdir, run_ids=None, verbose=False, show_hidden=False):
	outdir = os.path.abspath(outdir)
	utils.safe_mkdir(outdir)

	if not id and run_ids is None:
		utils.die("must specify a catalog id or a list of run ids")

	if id:
		record = catalog.select(id)
		if record is None:
			utils.die("no catalog entry found for id '%s'" % id)
		title = record.species if record.species else id
	else:
		title = "Agalma runs %s" % ','.join(run_ids)
	
	html = ["""
		<html>
		<head>
		<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
		<title>{0}</title>
		<style>{1}</style>
		{2}
		</head>
		<body>
		<h1><em>{0}</em></h1>""".format(title, report.stylesheet, report.javascript)]

	if id:
		html += catalog_header(record)

	toc = list()
	content = list()
	sections = (
		sanitize,
		insert_size,
		remove_rrna,
		assemble,
		postassemble,
		homologize,
		multalign)
	sections = OrderedDict([(section, list()) for section in sections])

	if run_ids:
		runs = map(diagnostics.lookup_run, run_ids)
	else:
		runs = diagnostics.lookup_runs(id)

	for run in runs:
		if (not show_hidden) and run.hidden > 0:
			utils.info("skipping hidden run %d" % run.run_id)
			continue
		links = list()
		# Try to generate a report for each section and add it to the content.
		run.pipelines = diagnostics.lookup_pipelines(run.run_id)
		for section, content in sections.iteritems():
			name = section.__name__
			if name in run.pipelines:
				run_report = str(section.Report(id, run.run_id, outdir, hlevel=3, verbose=verbose))
				if run_report:
					links.append("<li><a href=\"#{0}_{1}\">{0}</a></li>".format(name, run.run_id))
					content += [
						"<h3><a name=\"{0}_{1}\">Run {1}</a></h3>".format(name, run.run_id),
						run_report,
						"<p><a href=\"#toc\">Back to TOC</a></p>"]
		# Print a header with basic information about the run.
		toc_row = Template("""
			<tr class=\"minimal\">
				<td class=\"minimal\">
				<b>Run $run_id</b><br/><i>$name</i><br/>$timestamp<br/>$hostname
				</td>
				<td class=\"minimal\">
				<ul>$links</ul>
				</td>
				<td class=\"minimal\">
				<pre>$resources</pre>
				</td>
			</tr>""")
		run.links = ''.join(links)
		run.resources = profile_pipeline(run.run_id)
		toc.append(toc_row.safe_substitute(run))

	html.append("<h2><a name=\"toc\">Table of Contents</a></h2><table class=\"minimal\">")
	html += toc
	html.append("</table><hr>")
	for section, content in sections.iteritems():
		if content:
			html.append("<h2>{}</h2><p>{}</p>".format(section.__name__, section.__doc__))
			html += content
			html.append("<hr>")
	html.append("</body></html>")

	# Final output.
	with codecs.open(os.path.join(outdir, 'index.html'), 'w', 'utf-8') as f:
		f.write('\n'.join(line.replace('\t','') for line in html if line))

if __name__ == '__main__':
	parser = argparse.ArgumentParser(description="""
		Generates an HTML report of all runs in the agalma diagnostics database
		for the given agalma catalog ID (by default), or of only the specific
		list of RUN_IDS""")
	parser.add_argument('--outdir', '-o', default='./', help="""
  		write HTML output to OUTDIR [default: ./]""")
	parser.add_argument('--verbose', '-v', action='store_true', help="""
		include verbose output with additional plots and information, but
		with a longer report generation time [default: False]""")
	parser.add_argument('--hidden', action='store_true', help="""
		include runs that are marked as hidden [default: False]""")
	parser.add_argument('--id', '-i', metavar='ID', help="""
		catalog ID""")
	parser.add_argument('run_ids', metavar='RUN_ID', nargs='*', help="""
		include only the specified list of run IDs""")
	args = parser.parse_args()
	report_runs(args.id, args.outdir, args.run_ids, args.verbose, args.hidden)

