#!/usr/bin/env python
#
# Agalma - Tools for processing gene sequence data and automating workflows
# Copyright (c) 2012 Brown University. All rights reserved.
# 
# This file is part of Agalma.
# 
# Agalma is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Agalma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Agalma.  If not, see <http://www.gnu.org/licenses/>.

'''
treeprune

Prune gene trees.

For each gene tree generated in genetree, it prunes the tree to
include only one representative sequence per taxon when sequences
form a monophyloetic group (here called 'monophyly masking').  It then
prunes the monophyly-masked tree into maximally inclusive subtrees with
no more than one sequence per taxon (here called 'paralogy pruning').

'''

import glob
import os
import shutil

from dendropy import Tree

import database

from biolite import diagnostics
from biolite import report
from biolite import utils
from biolite import wrappers

from biolite.pipeline import Pipeline

pipe = Pipeline('treeprune', __doc__) 

# Command-line arguments
pipe.add_argument('tree_id', default=None, metavar='ID', help="""
	Use the trees from a previous genetree run.""")
pipe.add_argument('tree_dir', default=None, metavar='DIR', help="""
	Use an explicit path to a directory containing a 'RAxML_bestTree.*'
	file for each cluster.""")


# Pipeline stages
@pipe.stage
def setup_path(id, tree_id, tree_dir):
	"""Determine the path of the input genetree"""
	
	if tree_dir:
		tree_dir = os.path.abspath(tree_dir)
	elif tree_id:
		# Lookup values for a previous run
		values = diagnostics.lookup(tree_id, diagnostics.EXIT)
		try:
			tree_dir = values['raxml_dir']
		except KeyError:
			utils.die("No 'raxml_dir' entry found in run %s" % tree_id)
	else:
		utils.die(
			"No directory or run id provided: use --tree_dir or --tree_id")

	ingest('tree_dir')


@pipe.stage
def prune_trees(tree_dir, outdir):
	"""Prune each tree using monophyly masking and paralogy pruning"""

	prune_dir = os.path.join(outdir, 'pruned-trees')
	utils.safe_mkdir(prune_dir)

	# RAxML generates by default 4 files for each alignment; here we only need
	# the best trees
	for tree in glob.glob(os.path.join(tree_dir, 'RAxML_bestTree.*')):
		basename = os.path.basename(tree)
		diagnostics.prefix.append(basename)
		shutil.copy(tree, prune_dir)
		wrappers.TreePruning(os.path.join(prune_dir, basename))
		diagnostics.prefix.pop()

	ingest('prune_dir')


@pipe.stage
def parse_trees(prune_dir, _run_id):
	"""Parse the tips of each tree to create a cluster in the database"""

	hist = {}
	rows = []

	# Parse each paralogy pruned tree, and create a cluster from the tips
	cluster_id = 0
	for tree_path in glob.glob(os.path.join(prune_dir, '*.pp.*')):
		tree = Tree(stream=open(tree_path,'rU'),
					schema='newick',
					as_unrooted=True)
		nodes = tree.leaf_nodes()
		size = len(nodes)
		# Only keep trees with 3 or more tips
		if size >= 3:
			hist[size] = hist.get(size, 0) + 1
			for node in nodes:
				seq_id = node.get_node_str().partition('@')[-1]
				rows.append((_run_id, cluster_id, seq_id))
			cluster_id += 1

	# Write clusters to the database
	database.execute("BEGIN")
	database.execute("DELETE FROM homology WHERE run_id=?;", (_run_id,))
	homology_sql = """
		INSERT INTO homology (run_id, component_id, sequence_id)
		VALUES (?,?,?);"""
	for row in rows:
		database.execute(homology_sql, row)
	database.execute("COMMIT")

	utils.info("histogram of gene cluster sizes:")
	for k in sorted(hist):
		print "%d\t:\t%d" % (k, hist[k])

	diagnostics.log('histogram', hist)


if __name__ == "__main__":
	# Run the pipeline.
	pipe.run()
	# Push the local diagnostics to the global database.
	diagnostics.merge()

# Report. #
class Report(report.BaseReport):
	def init(self):
		self.name = pipe.name
		# Lookups
		self.lookup('outdir', diagnostics.INIT, 'outdir')
		# Generators

